package org.capcaval.ccoutils.lafabrique;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.List;

import org.capcaval.ccoutils.commandline.Command;
import org.capcaval.ccoutils.commandline.CommandParam;
import org.capcaval.ccoutils.common.CommandResult;
import org.capcaval.ccoutils.common.CommandResult.Type;
import org.capcaval.ccoutils.compiler.CompilerTools;
import org.capcaval.ccoutils.file.FileTools;
import org.capcaval.ccoutils.file.FileTools.FilePropertyEnum;
import org.capcaval.ccoutils.lafabrique.command.CommandCompile;
import org.capcaval.ccoutils.lafabrique.command.CommandEclipseProject;
import org.capcaval.ccoutils.lafabrique.command.CommandJar;
import org.capcaval.ccoutils.lafabrique.command.CommandNewApplication;
import org.capcaval.ccoutils.lafabrique.command.CommandNewProject;
import org.capcaval.ccoutils.lafabrique.command.CommandPack;
import org.capcaval.ccoutils.lang.ArrayTools;
import org.capcaval.ccoutils.lang.ClassTools;
import org.capcaval.ccoutils.lang.StringMultiLine;


public class LaFabriqueCommands {

	@Command(desc="Update eclipse project from ccOutils project.")
	public String updateEclipseProject( 
			@CommandParam(name="project name", desc="Define the name of the project to be updated.")
			String projectNameStr) {
		CommandResult cr1 = new CommandResult("Update Eclipse project named " + projectNameStr);

		// compile the project first
		AbstractProject project = this.retrieveProject(projectNameStr);

		if(project == null){
			StringMultiLine returnedMessage = new StringMultiLine(
					"Error project " + projectNameStr + " can not be found.",
					"                     the file 00_prj/prj/" + projectNameStr + ".java can not be found.");
			
			CommandResult crError = new CommandResult(returnedMessage.toString());
			
			return cr1.toString() + crError.toString();
		}
		
		CommandResult cr2 = CommandEclipseProject.updateEclipseProject(project);
		
		// by default do it one the current directory
		return cr1.toString() + cr2.toString();
	}

	@Command(desc="Create an application inside this project.")
	public String newApp( 
			@CommandParam(name="application full class name", desc="Define the name of the application to be created.")
			String fullClassName) {
		CommandResult cr1 = new CommandResult("New application " + fullClassName);

		// compile the default project first
		AbstractProject project = this.retrieveDefaultProject();

		if(project == null){
			StringMultiLine returnedMessage = new StringMultiLine(
					"Error default project can not be found.",
					"                     the file 00_prj/prj/DefaultProject.java can not be found or is invalid.");
			
			CommandResult crError = new CommandResult(returnedMessage.toString());
			
			return cr1.toString() + crError.toString();
		}
				
		CommandResult cr2 = CommandNewApplication.newApplication(project, fullClassName);
		
		// by default do it one the current directory
		return cr1.toString() + cr2.toString();
	}
	
	@Command
	public String newProject(
			@CommandParam(name="project name", desc="Define the name of the project.")
			String name, 
			@CommandParam(name="project path", desc="Define the path of the project.")
			String pathStr) {
		// get the default configuration of project
		DefaultConfigProject defaultConfigProject = new DefaultConfigProject();

		Path projectPath = Paths.get(pathStr, name);

		// clean the directory
		try {
			if (projectPath.toFile().exists() == true) {
				FileTools.deleteFile(projectPath);
			}
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		CommandResult cr = CommandNewProject.newProject(name, projectPath, defaultConfigProject);
				
		return cr.toString();
	}

	protected AbstractProject compileProject(Path path) {
		AbstractProject proj=null;
		try {
			proj = CompilerTools.compile(AbstractProject.class, path.toString() + "/00_prj", "prj/Project.java", "./10_bin");
			// set the current path of the project
			proj.rootProjectDirPath = path.toAbsolutePath();
		} catch (Exception e) {
			e.printStackTrace();
		}

		return proj;
	}


	@Command(desc="build project")
	public String build(
			@CommandParam(name="project name", desc="Name of the project to be built.")
			String projectStr) {
		StringMultiLine returnedMessage = new StringMultiLine("[laFabrique] INFO  : Build start " + projectStr  + " project.");

		// compile the project first
		AbstractProject proj = this.retrieveProject(projectStr);

		if(proj == null){
			returnedMessage.addLine("[laFabrique] ERROR : Error project " + projectStr + " can not be found.");
			returnedMessage.addLine("                     the file 00_prj/prj/" + projectStr + ".java can not be found.");
			return returnedMessage.toString();
		}
		
		try {
			returnedMessage.addLine("[laFabrique] INFO  : Compile all classes");
			// Now that the project is there let's build it
			CommandResult cr = CommandCompile.compile(proj);
			returnedMessage.addLine( cr.getReturnMessage());

			if ((proj.jar.name != null)&&(cr.getType() != Type.ERROR)) {
				returnedMessage.addLine("[laFabrique] INFO  : Build the jar");
				cr = CommandJar.makeJar(proj);
				returnedMessage.addLine(cr.getReturnMessage());
				
				// copy the new jar inside the sample directory
				//copyFile(returnedMessage, proj.productionDirPath.toString() + "/" + proj.jar.name, "30_sample/02_lib");
				Path src = Paths.get(proj.productionDirPath.toString() + "/" + proj.jar.name);
				Path dst = Paths.get("30_sample/02_lib/"+ proj.jar.name);
				FileTools.copy( src, dst);
				returnedMessage.addLine("[laFabrique] INFO  : copy the jar to " + dst.toString());
				dst = Paths.get("02_lib/"+ proj.jar.name);
				FileTools.copy( src, dst);
				returnedMessage.addLine("[laFabrique] INFO  : copy the jar to " + dst.toString());
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}

		return returnedMessage.toString();
	}
	@Command(desc="pack project")
	public String pack(
			@CommandParam(name="project name", desc="Name of the project to be packed.")
			String projectStr
			){
		StringMultiLine returnedMessage = new StringMultiLine("[laFabrique] INFO  : Start to pack project " + projectStr  + " project.");
		
		// compile the project first
		AbstractProject proj = this.retrieveProject(projectStr);
		
		CommandResult result = CommandPack.pack(proj);
		returnedMessage.addLine(result.toString());
		
		return returnedMessage.toString();
	}
	
	
	public AbstractProject retrieveProject(String projectName){
		String filePath = "prj/" + projectName + ".java";
		
		// if the project do not exist return null
		if(FileTools.isFileExist("00_prj/" + filePath) == false){
			return null;
		}
		
		// compile the project first
		AbstractProject proj = null;
		try {
			proj = CompilerTools.compile(AbstractProject.class, "00_prj", filePath, "10_bin");
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return proj;
	}

	public AbstractProject retrieveDefaultProject(){
		String filePath = "prj/DefaultProject.java";
		
		// if the project do not exist return null
		if(FileTools.isFileExist("00_prj/" + filePath) == false){
			return null;
		}
		
		// compile the project first
		AbstractDefaultProject defaultProj = null;
		try {
			defaultProj = CompilerTools.compile(AbstractDefaultProject.class, "00_prj", filePath, "10_bin");
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return retrieveProject(defaultProj.defaultProject.getSimpleName());
	}
	
}
